/** \file
*/

/*
* ----------------------------------------------------------------------------
* "THE BEER-WARE LICENSE":
* tomasz.o.ostrowski <that funny sign> gmail.com wrote this file.
* As long as you retain this notice you can do whatever you want with
* this stuff. If we meet some day, and you thinkthis stuff is worth it,
* you can buy me a beer in return
* Tomasz Ostrowski
* ----------------------------------------------------------------------------
*/

#include <vcl.h>
#pragma hdrstop

#include "frmMainUnit.h"
#include "about.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TfrmMain *frmMain;

/** \brief Callback for WaveIn Device
*/
void CALLBACK WaveInProc(HWAVEIN waveOut, UINT uMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2)
{
    //The MM_WIM_DATA message is sent to a window when waveform-audio data
    //is present in the input buffer and the buffer is being returned to
    //the application. The message can be sent either when the buffer is full
    //or after the waveInReset function is called.
    //wParam = (WPARAM) hInputDev
    //lParam = (LONG) lpwvhdr
    if ( uMsg == MM_WIM_DATA )
    {
        frmMain->ProcessSamples();
    }
}
//---------------------------------------------------------------------------
__fastcall TfrmMain::TfrmMain(TComponent* Owner)
        : TForm(Owner)
{
}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::ProcessSamples()
{
//
//  Prepare data buffer
//
    if ( bRecordingStarted )
    {
        WAVEHDR *in = &WaveHeaders[iBuffIdx];
		//The waveInUnprepareHeader function cleans up the preparation performed
        //by the waveInPrepareHeader function. This function must be called
        //after the device driver fills a buffer and returns it to
        //the application. You must call this function before freeing the buffer.
        ::waveInUnprepareHeader( hWaveIn, in, sizeof(WAVEHDR) );

		// process samples
		timethread->Process(pboxGraph,
			lblLeft->Top + lblLeft->Height/2 - 1,
			lblRight->Top + lblRight->Height/2 - 1,
			(short*)(in->lpData));

		// next buffer to send
        iBuffIdx++;
        if ( iBuffIdx == iBuffersCount )
            iBuffIdx=0;
        SendHeader->dwBufferLength=WaveBufSize;
        SendHeader->dwFlags=0;
        SendHeader->dwLoops=0;
        ::waveInPrepareHeader( hWaveIn, SendHeader, sizeof(WAVEHDR) );
        ::waveInAddBuffer( hWaveIn, SendHeader, sizeof(WAVEHDR));
        if ( SendHeader == &WaveHeaders[iBuffersCount-1] )
            SendHeader = &WaveHeaders[0];
        else
            SendHeader++;
    }
}

void __fastcall TfrmMain::FormCreate(TObject *Sender)
{
	GetWaveInDevices();

	iBuffersCount = 2;	// uses double-buffering
    iSamplesPerChannel = 1024; ///< total samples count (per channel)
    iBuffIdx = 0;
    waveFormat.wFormatTag = WAVE_FORMAT_PCM;
    waveFormat.nChannels = 2;
    waveFormat.nSamplesPerSec = 44100;
    waveFormat.wBitsPerSample = 16;
    waveFormat.nBlockAlign = waveFormat.wBitsPerSample/8*waveFormat.nChannels;
    waveFormat.nAvgBytesPerSec = waveFormat.nBlockAlign*waveFormat.nSamplesPerSec;
    waveFormat.cbSize = 0;
    timethread = new DrawThread( true, iSamplesPerChannel );

	//  Audio buffers
	WaveHeaders = new WAVEHDR[iBuffersCount];
    WaveBufSize = iSamplesPerChannel * (waveFormat.wBitsPerSample/8) * waveFormat.nChannels;
    for ( int i = 0; i<iBuffersCount; i++ )
    {
        WaveHeaders[i].dwBufferLength = WaveBufSize;
        WaveHeaders[i].lpData = (char *)VirtualAlloc( 0, WaveBufSize, MEM_COMMIT, PAGE_READWRITE );
    }
    if ( waveInOpen( &hWaveIn, WAVE_MAPPER, &waveFormat, (DWORD)WaveInProc, (DWORD)this, CALLBACK_FUNCTION ) == MMSYSERR_NOERROR)
    {
        for ( int i = 0; i<iBuffersCount/2; i++ )
        {
            WaveHeaders[i].dwBufferLength=WaveBufSize;
            WaveHeaders[i].dwFlags=0;
            WaveHeaders[i].dwLoops=0;
            if ( waveInPrepareHeader( hWaveIn, &WaveHeaders[i], sizeof(WAVEHDR))== MMSYSERR_NOERROR)
                if ( waveInAddBuffer( hWaveIn, &WaveHeaders[i], sizeof(WAVEHDR))!= MMSYSERR_NOERROR)
                {
                    Application->MessageBox("Error opening audio device."
                        "Check if device is capable of recording with selected parameters or isn't busy.\n",
                        this->Caption.c_str(), MB_ICONEXCLAMATION);
                    Application->Terminate();
                }
        }
        if ( waveInStart( hWaveIn ) == MMSYSERR_NOERROR )
        {
            SendHeader = &WaveHeaders[iBuffersCount/2];
            iBuffIdx = 0;
            bRecordingStarted = true;
        }
        else
        {
            Application->MessageBox( "Error", "Multimedia Error", MB_OK);
            Application->Terminate();
        }
    }
    else
    {
        Application->MessageBox( "Error", "Multimedia Error", MB_OK);
        Application->Terminate();
    }
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::FormDestroy(TObject *Sender)
{
    bRecordingStarted = false;
    timethread->Terminate();

	// close wave device
    MMRESULT res=waveInReset( hWaveIn );
    for ( int i = 0; i<iBuffersCount; i++ )
    {
        res = ::waveInUnprepareHeader( hWaveIn, &WaveHeaders[i], sizeof(WAVEHDR) );
        while ( res == WAVERR_STILLPLAYING )
        {
            // The buffer pointed to by the pwh parameter is still in the queue.
            Sleep( 50 );
            res = ::waveInUnprepareHeader( hWaveIn, &WaveHeaders[i], sizeof(WAVEHDR) );
        }
    }
    ::waveInClose( hWaveIn );

    delete timethread;

    for ( int i=0; i<iBuffersCount; i++ )
        VirtualFree( WaveHeaders[i].lpData, 0, MEM_RELEASE );
    delete [] WaveHeaders;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::miSelectInputClick(TObject *Sender)
{
	// deselect all menu items
	for (int i = miInput->Count - 1; i>=0; --i)
	{
		miInput->Items[i]->Checked = false;
	}
	TMenuItem *clicked;
	clicked = dynamic_cast<TMenuItem*>(Sender);
	clicked->Checked = true;
}
//---------------------------------------------------------------------------

/** Detect available WAVE IN devices
*/
void TfrmMain::GetWaveInDevices(void)
{
	WAVEINCAPS     wic;

	TMenuItem *mi = new TMenuItem(miInput);
	mi->Caption = "Default (WAVE_MAPPER)";
	mi->Tag = 999;
	mi->OnClick = miSelectInputClick;
	mi->Checked = true;
	mi->RadioItem = true;
	miInput->Insert(0, mi);

	// Get the number of Digital Audio In devices in this computer
	int iNumDevs = waveInGetNumDevs();
	for (int i = 0; i < iNumDevs; i++)
	{
		// Get info about the next device
		if (!waveInGetDevCaps(i, &wic, sizeof(WAVEINCAPS)))
		{
			// Add device as menu item
			mi = new TMenuItem(miInput);
			mi->Caption = wic.szPname;
			mi->Tag = i;
			mi->OnClick = miSelectInputClick;
			mi->Checked = false;
			mi->RadioItem = true;
			miInput->Insert(i+1, mi);
		}
	}
}
void __fastcall TfrmMain::FormResize(TObject *Sender)
{
    static int iLastPboxHeight = pboxGraph->Height;
	pboxGraph->Width = this->Width - 58;
	pboxGraph->Height = this->Height - 86;
	imgMouse->Top = this->Height - 89;
	lblLeft->Top = (float)(lblLeft->Top - pboxGraph->Top)/(float)iLastPboxHeight *
	 	pboxGraph->Height + pboxGraph->Top;
	lblRight->Top = (float)(lblRight->Top - pboxGraph->Top)/(float)iLastPboxHeight *
		pboxGraph->Height + pboxGraph->Top;
	iLastPboxHeight = pboxGraph->Height;
	pboxGraph->Update();
}
//---------------------------------------------------------------------------


void __fastcall TfrmMain::FormMouseDown(TObject *Sender, TMouseButton Button,
      TShiftState Shift, int X, int Y)
{
	if (Sender != frmMain)
	{
		TPoint P;
		GetCursorPos(&P);
		P = frmMain->ScreenToClient(P);
		X = P.x;
		Y = P.y;
	}
	if (X>10 && X < pboxGraph->Left)
	{
		if (Y>pboxGraph->Top + 15 && Y<pboxGraph->Top + pboxGraph->Height - 15)
        {
			if (Button == mbLeft)
			{
				lblLeft->Top = Y - lblLeft->Height/2;
			}
			else if (Button == mbRight)
				 {
					 lblRight->Top = Y - lblRight->Height/2;
				 }
		}
	}
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::About1Click(TObject *Sender)
{
	AboutBox->ShowModal();	
}
//---------------------------------------------------------------------------

